import numpy as np
import pandas as pd
from metabotk.utils import validate_dataframe


class OutlierHandler:
    """
    Class for detecting outliers using the Interquartile Range (IQR) method.
    """

    def detect_outliers(self, data, threshold):
        """
        Detect outlier values in a single-column numerical array.

        Parameters:
        - data: single-column numerical array
        - threshold: a factor that determines the range from the IQR

        Returns:
        - Boolean array indicating outliers (True) and non-outliers (False)
        """
        if len(data) == 0:
            raise ValueError("Input is empty.")
        if isinstance(data, pd.DataFrame):
            raise TypeError("DataFrame input is not supported.")

        median = np.nanmedian(data)
        q1 = np.nanquantile(data, 0.25)
        q3 = np.nanquantile(data, 0.75)
        iqr = q3 - q1
        cutoff_lower = median - (threshold * iqr)
        cutoff_upper = median + (threshold * iqr)
        is_outlier = (data < cutoff_lower) | (data > cutoff_upper)
        return is_outlier

    def get_outliers_matrix(self, data_frame, threshold, axis=0):
        """
        Get a matrix indicating outliers in each row or column of a dataframe.

        Parameters:
        - data_frame: pandas DataFrame containing only numeric values
        - threshold: a factor that determines the range from the IQR
        - axis: {0 or ‘index’, apply to each column, 1 or ‘columns’, apply to each row}, default 0

        Returns:
        - pandas DataFrame indicating outliers (True) and non-outliers (False)
        """
        validate_dataframe(data_frame)
        matrix = data_frame.apply(
            lambda x: self.detect_outliers(x, threshold), axis=axis
        )
        return matrix

    def count_outliers(self, data_frame, threshold, axis=0):
        """
        Count number of outlier values in each row or column of a dataframe.

        Parameters:
        - data_frame: pandas DataFrame containing only numeric values
        - threshold: a factor that determines the range from the IQR
        - axis: {0 or ‘index’, apply to each column, 1 or ‘columns’, apply to each row}, default 0

        Returns:
        - pandas Series with the row/column index and the number of outliers
        """
        validate_dataframe(data_frame)
        outliers_matrix = self.get_outliers_matrix(data_frame, threshold)
        outlier_counts = outliers_matrix.sum(axis=axis)
        return outlier_counts

    def remove_outliers(self, data_frame, threshold, axis=0):
        """
        Replace outlier values with NAs in a dataframe, column-wise or row-wise.

        Parameters:
        - data_frame: pandas DataFrame containing only numeric values
        - threshold: a factor that determines the range from the IQR
        - axis: {0 or ‘index’, apply to each column, 1 or ‘columns’, apply to each row}, default 0

        Returns:
        - pandas DataFrame where the outlier values are replaced by NAs
        """
        validate_dataframe(data_frame)
        outliers = self.get_outliers_matrix(data_frame, threshold, axis=axis)
        data_frame_without_outliers = data_frame.where(~outliers, np.nan)
        return data_frame_without_outliers



import miceforest as mf
import numpy as np
import pandas as pd
from sklearn.utils import check_random_state
from metabotk.statistics_handler import get_top_n_correlations


class ImputationHandler:
    """
    Class for performing missing data imputation on metabolomics data.

    This class provides a simple interface to perform missing value imputation on the data using various approaches.
    For now only miceforest is included.

    Attributes:
        stats (StatisticsHandler): StatisticsHandler instance containing the data.
    """

    def __init__(
        self,
        dataset_manager,
    ):
        """
        Initialize the class.

        Parameters:
            statistics_handler (StatisticsHandler): StatisticsHandler instance containing the data.
            n_imputed_datasets (int): Number of imputed datasets to generate (default: 5).
            n_iterations (int): Number of iterations of MICE to perform (default: 5).
            random_state (int): Random seed (default: None).
        """
        self._dataset_manager = dataset_manager

    def miceforest(
        self,
        n_correlated_metabolites,
        n_imputed_datasets=5,
        n_iterations=5,
        random_state=None,
        get_kernel=False,
    ):
        """
        Perform missing data imputation using MICE (Mixed-effects Imputation by Chained Equations).

        Parameters:
            n_correlated_metabolites (int): Number of metabolites to use for correlated imputation.

        Returns:
            dict: Dictionary with imputed datasets, keys are integers starting from 1.
        """
        corrs = get_top_n_correlations(
            data_frame=self._dataset_manager.data, n=n_correlated_metabolites
        )
        corrs_dict = {}
        for name, group in corrs.groupby(by="id"):
            corrs_dict[name] = group["correlated_ids"].tolist()
        kds = mf.ImputationKernel(
            data=self._dataset_manager.data,
            datasets=n_imputed_datasets,
            variable_schema=corrs_dict,
            save_all_iterations=True,
            random_state=random_state,
            train_nonmissing=False,
            mean_match_scheme=mf.mean_match_shap,
        )
        kds.mice(
            iterations=n_iterations,
            verbose=True,
        )

        imputed = {
            dataset + 1: kds.complete_data(dataset=dataset)
            for dataset in range(n_imputed_datasets)
        }
        if get_kernel == True:
            return kds, imputed
        else:
            return imputed



from .outliers_handler import OutlierHandler
from .missing_handler import MissingDataHandler
from .statistics_handler import StatisticsHandler
from .dataset_manager import DatasetManager
from .models_handler import ModelsHandler
from .visualization_handler import Visualization
from .dimensionality_reduction import DimensionalityReduction
from .imputation import ImputationHandler
from .feature_selection import FeatureSelection

from ._version import __version__
from .interface import MetaboTK

__all__ = [
    "__version__",
]



import pandas as pd
import numpy as np
from pathlib import Path


def validate_dataframe(data_frame):
    """
    Validates that the input is a pandas DataFrame.

    Parameters
    ----------
    data_frame : any
        Input to validate.

    Raises
    ------
    TypeError
        If the input is not a pandas DataFrame.
    """
    if not isinstance(data_frame, pd.DataFrame):
        raise TypeError("Data must be a pandas DataFrame.")


def ensure_numeric_data(data):
    """
    Ensure that the input data is numeric.

    This function checks that the input data is a non-empty NumPy array
    containing only numeric values. If the data is not numeric or empty,
    a TypeError is raised.

    Parameters
    ----------
    data : list, array, or Series
        Input data to validate.

    Returns
    -------
    np.array
        Input data converted to NumPy array.

    Raises
    ------
    ValueError
        If the input data is empty.
    TypeError
        If the input data is not numeric.
    """
    if len(data) == 0:
        raise ValueError("Empty data")
    data = np.array(data)  # convert to np array
    if not np.issubdtype(data.dtype, np.number):
        raise TypeError("Data must contain only numeric values")
    return data


def parse_input(input_data):
    """
    Parse input data as pandas dataframe or as file path to TSV or CSV file

    This function allows users to provide input data as a pandas DataFrame or
    as a file path to a TSV or CSV file. If the input is a DataFrame, it is
    returned as is. If the input is a file path, the function loads the data
    from the file and returns it as a DataFrame.

    Parameters
    ----------
    input_data : pandas.DataFrame or str
        Input data to be parsed.

    Returns
    -------
    pandas.DataFrame
        Input data as a pandas DataFrame.

    Raises
    ------
    TypeError
        If the input is not a pandas DataFrame or a file path.
    """
    if isinstance(input_data, pd.DataFrame):
        data = input_data.reset_index(drop=True)
        return data
    elif isinstance(input_data, str):
        if input_data.endswith((".tsv", ".csv")):
            if input_data.endswith(".tsv"):
                data = pd.read_table(input_data, sep="\t").reset_index(drop=True)
            elif input_data.endswith(".csv"):
                data = pd.read_csv(input_data).reset_index(drop=True)
            return data
        else:
            raise TypeError(
                "Input should be a Pandas DataFrame or a file path to a TSV or CSV file."
            )
    else:
        raise TypeError(
            "Input should be a Pandas DataFrame or a file path to a TSV or CSV file."
        )


def create_directory(directory_path):
    """
    Creates a directory at the given path.

    This function creates a directory at the given path and all parent directories
    if they do not already exist. If the directory already exists, nothing is done.

    Parameters
    ----------
    directory_path : str
        Path of the directory to be created.
    """
    Path(directory_path).mkdir(parents=True, exist_ok=True)


def is_int(s):
    try:
        int(s)
        return True
    except ValueError:
        return False



import numpy as np
import statsmodels.formula.api as smf
import dill
from metabotk.utils import create_directory


class ModelsHandler:
    """
    Class for fitting models to the data and obtaining residuals

    The class takes a data manager as input, and fits a linear model using the
    formula specified in the constructor for each metabolite. The residuals
    are then extracted from the fitted models and returned as a pandas DataFrame.

    If a path to a directory is provided, the fitted models will be saved as
    pickle files in that directory.

    Attributes:
        data_manager (DataManager): DataManager instance containing the data
        formula (str): formula used to fit the linear model
        merged (DataFrame): merged dataframe of sample metadata and data
        residuals (DataFrame): dataframe of residuals, initially full of NaNs
    """

    def __init__(self, data_manager) -> None:
        """
        Initialize the class.

        Parameters:
            data_manager (DataManager): DataManager instance containing the data
            formula (str): formula used to fit the linear model
            models_path (str): path to directory where models will be saved
        """
        self.data_manager = data_manager
        self.merged = self.data_manager.merge_sample_metadata_data()
        self.residuals = self.data_manager.data.copy()
        self.residuals.loc[:] = np.nan

    def fit_linear_model(self, metabolite, formula):
        """
        Fit a linear model using the formula specified in the constructor.

        Parameters:
            metabolite (str): name of metabolite to fit model for

        Returns:
            residuals (Series): residuals from the fitted model
            model (RegressionResults): fitted model
        """
        model = smf.ols(f"Q('{metabolite}') ~ {formula}", self.merged)
        fitted_model = model.fit()
        residuals = fitted_model.resid
        return residuals, model

    def get_linear_model_residuals(self, formula, models_path=None):
        """
        Fit a linear model for each metabolite and extract residuals.

        Parameters:
            models_path (str): path to directory where models will be saved

        Returns:
            residuals (DataFrame): dataframe of residuals for all metabolites
        """
        if models_path:
            create_directory(models_path)
        for metabolite in self.data_manager.metabolites:
            residuals, model = self.fit_linear_model(metabolite, formula)
            self.residuals[metabolite] = residuals
            if models_path:
                with open(f"{models_path}/{metabolite}.pickle", "wb") as handle:
                    dill.dump(model, handle)
        return self.residuals



"""
Module for parsing input data
"""

from typing import Union
import pandas as pd
from metabotk.utils import parse_input


class MetabolonCDT:
    """
    A class for reading Metabolon Client Data Table files.

    Attributes:
    - data_key_and_explanation: DataFrame containing data key and explanation.
    - sample_metadata: DataFrame containing sample metadata (required).
    - chemical_annotation: DataFrame containing chemical annotation (required).
    - peak_area_data: DataFrame containing peak area data and a column with the sample names. (required)
    - batch_normalized_data: DataFrame containing batch-normalized data and a column with the sample names. (optional)
    - batch_normalized_imputed_data: DataFrame containing batch-norm imputed data and a column with the sample names. (optional)
    - log_transformed_data: DataFrame containing log-transformed data and a column with the sample names.(optional)
    """

    def __init__(self) -> None:
        """
        Initialize the class.
        """
        self.data_key_and_explanation = None
        self.sample_metadata = None
        self.chemical_annotation = None
        self.generic_data = None
        self.peak_area_data = None
        self.batch_normalized_data = None
        self.batch_normalized_imputed_data = None
        self.log_transformed_data = None
        self.additional_data = {}

    def import_excel(
        self,
        file_path: str,
        sample_metadata: str = "Sample Meta Data",
        chemical_annotation: str = "Chemical Annotation",
        data: str = "Peak Area Data",
    ) -> None:
        """
        Read the Metabolon Client Data Table Excel file and assign its sheets to class attributes.

        Args:
            file_path (str): The file path to the Metabolon Excel file.

        Raises:
            ValueError: If any required sheet ('Sample Meta Data', 'Chemical Annotation',
            'Peak Area Data') is missing from the Excel file.
        """
        # Read Excel file
        sheets = pd.read_excel(file_path, sheet_name=None)

        # Check if required sheets are missing
        missing_sheets = [
            sheet_name
            for sheet_name in [
                sample_metadata,
                chemical_annotation,
                data,
            ]
            if sheet_name not in sheets
        ]
        if missing_sheets:
            raise ValueError(
                "The following required sheets are missing: {}".format(missing_sheets)
            )

        # Assign sheets to attributes
        self.data_key_and_explanation = sheets.pop("Data Key & Explanation", None)
        self.sample_metadata = sheets.pop(sample_metadata)
        self.chemical_annotation = sheets.pop(chemical_annotation)
        self.generic_data = sheets.pop(data)
        self.peak_area_data = sheets.pop("Peak Area Data", None)
        self.batch_normalized_data = sheets.pop("Batch-normalized Data", None)
        self.batch_normalized_imputed_data = sheets.pop("Batch-norm Imputed Data", None)
        self.log_transformed_data = sheets.pop("Log Transformed Data", None)
        self.additional_data = sheets

    def import_tables(
        self,
        sample_metadata: Union[pd.DataFrame, str],
        chemical_annotation: Union[pd.DataFrame, str],
        peak_area_data: Union[pd.DataFrame, str] = None,
        batch_normalized_data: Union[pd.DataFrame, str] = None,
        batch_normalized_imputed_data: Union[pd.DataFrame, str] = None,
        log_transformed_data: Union[pd.DataFrame, str] = None,
        generic_data: Union[pd.DataFrame, str] = None,
    ) -> None:
        """
        Read flat tables for sample metadata, chemical annotation, and metabolite data.

        Args:
            sample_metadata: DataFrame or path of file containing sample metadata (required).
            chemical_annotation: DataFrame or path of file containing chemical annotation (required).
            peak_area_data: DataFrame or path of file containing metabolite data (required).

        Raises:
            ValueError: If any required sheet ('Sample Meta Data', 'Chemical Annotation',
             'Peak Area Data') is missing from the Excel file.
        """
        # TODO: handling of other metabolite data such as batch normalized, log transformed etc
        # instead of generic metabolite_data
        # Read sample metadata
        self.sample_metadata = parse_input(sample_metadata)
        # Read chemical annotation
        self.chemical_annotation = parse_input(chemical_annotation)
        # Read peak area data
        try:
            self.peak_area_data = parse_input(peak_area_data)
        except TypeError:
            self.peak_area_data = None
        # Read batch normalized data
        try:
            self.batch_normalized_data = parse_input(batch_normalized_data)
        except TypeError:
            self.batch_normalized_data = None
        # Read batch normalized imputed data
        try:
            self.batch_normalized_imputed_data = parse_input(
                batch_normalized_imputed_data
            )
        except TypeError:
            self.batch_normalized_imputed_data = None
        # Read log transformed data
        try:
            self.log_transformed_data = parse_input(log_transformed_data)
        except TypeError:
            self.log_transformed_data = None
        # Read generic data
        try:
            self.generic_data = parse_input(generic_data)
        except TypeError:
            self.generic_data = None



"""
This module provides feature selection methods using scikit-learn.

Currently implemented methods:

* Boruta feature selection using the BorutaPy algorithm from boruta

"""

import dill
from sklearn.ensemble import RandomForestClassifier
from sklearn.model_selection import StratifiedKFold, KFold
from boruta_py_versioned import BorutaPy
import pandas as pd
from metabotk.utils import create_directory


class FeatureSelection:
    """
    Class for feature selection

    Attributes:
        data_manager (DatasetManager): DatasetManager object with data to be selected
        threads (int): Number of threads to use for parallel processing
        random_state (int): Random state for reproducibility
        group (str): Column name containing sample class information
    """

    def __init__(self, data_manager):
        """
        Initiate the class

        Args:
            data_manager (DatasetManager): DatasetManager object with data to be selected
        """
        self.data_manager = data_manager

    def setup_random_forest(self, random_state, threads, max_depth, class_weight):
        """
        Setup random forest

        Args:
            random_state (int): Random state for reproducibility
            max_depth (int): Max depth of the tree
            class_weight (dict): Dictionary with class weights

        Returns:
            RandomForestClassifier: Random Forest classifier
        """
        rf = RandomForestClassifier(
            n_jobs=threads,
            class_weight=class_weight,
            max_depth=max_depth,
            random_state=random_state,
        )
        return rf

    def boruta(
        self,
        y_column,
        threads=1,
        random_state=42,
        max_depth=None,
        class_weight="balanced",
        n_estimators="auto",
        alpha=0.01,
        max_iterations=1000,
        output_dir=None,
    ):
        """
        Boruta feature selection

        Performs Boruta feature selection using the BorutaPy algorithm.

        Args:
            max_depth (int): Max depth of the trees. If None, then no
                limit is set.
            class_weight (dict): Dictionary with class weights.
            n_estimators (int): Number of estimators for the Random Forest.
                If "auto", then the best value will be chosen.
            alpha (float): Confidence level for selecting features.
            max_iterations (int): Max iterations for the Boruta algorithm.
            random_state (int): Random state for reproducibility.
            get_model (bool): Return the Boruta model.

        Returns:
            tuple[DataFrame, DataFrame, BorutaPy]: Tuple with ranking and
                importance history. The ranking is a DataFrame with the
                metabolite as index and the ranking as values. The
                importance history is a DataFrame with the iteration
                number as rows and the metabolite as columns. If get_model
                is True, then the Boruta model is returned as well.
        """
        if not y_column:
            raise ValueError("y value must be specified")
        X = self.data_manager.data.values

        y = self.data_manager.sample_metadata[y_column].values

        rf = self.setup_random_forest(
            random_state=random_state,
            threads=threads,
            max_depth=max_depth,
            class_weight=class_weight,
        )
        feat_selector = BorutaPy(
            rf,
            n_estimators=n_estimators,
            alpha=alpha,
            verbose=2,
            random_state=random_state,
            max_iter=max_iterations,
        )
        feat_selector.fit(X, y)
        print(feat_selector.importance_history_)
        importance_history = pd.DataFrame(
            feat_selector.importance_history_,
            index=range(len(feat_selector.importance_history_)),
        )
        importance_history.columns = self.data_manager.metabolites
        importance_history.index.name = "iteration"
        importance_history = importance_history.reset_index()

        ranking = pd.DataFrame(
            feat_selector.ranking_,
            columns=["rank"],
            index=self.data_manager.metabolites,
        )
        ranking.index.name = "metabolite"
        ranking = ranking.reset_index()

        if output_dir:
            with open(f"{output_dir}/boruta_model.pickle", "wb") as f:
                dill.dump(feat_selector, f)
            importance_history.to_csv(
                f"{output_dir}/importance_history.tsv", sep="\t", index=False
            )

            ranking.to_csv(f"{output_dir}/ranking.tsv", sep="\t", index=False)
            return ranking
        else:
            return ranking

    def stratified_kfold(
        self,
        n_splits: int,
        stratification_column: str,
        output_dir: str,
    ) -> dict:
        """Split the dataset using a stratified approach for cross-validation.

        This function splits the dataset into n_splits folds using a stratified
        approach where the samples are split based on the values in the
        stratification_column in the sample metadata. The dataframes are saved
        to output_dir with filenames in the format of <fold_number>_train.tsv.
        The function returns a dictionary where the keys are the fold number
        and the values are the train dataframes.

        Args:
            n_splits: Number of splits to perform.
            output_dir: Directory to save dataframes to.
            stratification_column: Column in sample metadata to use for
                stratification.
        Returns:
            Dictionary with keys as fold numbers and values as the train
            dataframes.
        """
        create_directory(output_dir)
        X = self.data_manager.data
        split_train = {}
        split_val = {}
        y = self.data_manager.sample_metadata[stratification_column]
        skf = StratifiedKFold(n_splits=int(n_splits))
        skf_splits = skf.split(X, y)
        for fold, (train_idx, val_idx) in enumerate(skf_splits):
            foldname = fold + 1
            train = X.iloc[train_idx]
            val = X.iloc[val_idx]
            train.to_csv(f"{output_dir}/{foldname}_train.tsv", sep="\t")
            val.to_csv(f"{output_dir}/{foldname}_val.tsv", sep="\t")
            split_train[foldname] = train
            split_val[foldname] = val

        return split_train, split_val



from typing import Union, Optional, Dict
import pandas as pd
from metabotk.providers_handler import MetabolonCDT
from metabotk.utils import parse_input
import warnings


class DatasetManager:
    """
    Class for working with metabolomics data.

    This class provides an interface for loading and processing
    metabolomics data from different sources.

    Attributes:
        _data_provider (str): name of the data provider
        parser (object): parser for the data provider
        _sample_id_column (str): name of the sample id column
    """

    def __init__(
        self,
        data_provider: str = "metabolon",
        sample_id_column: str = None,
        metabolite_id_column: str = "CHEM_ID",
    ) -> None:
        """
        Initialize the class.

        Parameters:
            data_provider (str): name of the data provider
            sample_id_column (str): name of the sample id column
            metabolite_id_column (str): name of the metabolite id column
        """
        self._data_provider = data_provider.lower()
        if self._data_provider == "metabolon":
            self.parser = MetabolonCDT()
        else:
            raise NotImplementedError("This data provider is not supported yet")
        self._sample_id_column = sample_id_column
        self.sample_metadata = pd.DataFrame()
        # list of sample names
        self.samples = []
        self._metabolite_id_column = metabolite_id_column
        self.chemical_annotation = pd.DataFrame()
        # list of metabolite IDs
        self.metabolites = []
        # dataframe with sample x metabolite data
        self.data = pd.DataFrame()

    ###
    # Functions to import, setup and write data and metadata
    ###

    def _setup_data(self, chemical_annotation, sample_metadata, data) -> None:
        """
        Setup the class with chemical annotation, sample metadata, and data.

        This function loads the data and sets up the class instance.
        It checks that the sample ID column is found in the data and the
        metabolite ID column is found in the chemical annotation.
        If any of these checks fail, a ValueError is raised.

        Parameters:
            chemical_annotation (DataFrame or str): DataFrame or path of file containing chemical annotation.
            sample_metadata (DataFrame or str): DataFrame or path of file containing sample metadata.
            data (DataFrame or str): DataFrame containing or path of file the main data.

        Raises:
            ValueError: If the sample ID column is not found in the data or the metabolite ID column is not found in chemical annotation.
        """
        try:
            # parse input data
            parsed_chemical_annotation = parse_input(chemical_annotation)
            parsed_sample_metadata = parse_input(sample_metadata)
            parsed_data = parse_input(data)
            if not self._sample_id_column:
                if parsed_data.columns[0] in parsed_sample_metadata.columns:
                    self._sample_id_column = parsed_data.columns[0]
                else:
                    raise ValueError(
                        "No sample ID column found in data with correspondence in sample metadata"
                    )
            # check that sample ID column is found in data
            if (self._sample_id_column in parsed_data.columns) and (
                self._sample_id_column in parsed_sample_metadata.columns
            ):
                # set metadata and data
                self.sample_metadata = parsed_sample_metadata
                if self.sample_metadata is None:
                    raise ValueError("Sample metadata is not properly initialized.")
                if self.sample_metadata[self._sample_id_column].duplicated().any():
                    warnings.warn(
                        "Warning: there are duplicate values in the chosen sample column.\
                        Consider choosing another column or renaming the duplicated samples"
                    )
                self.sample_metadata[self._sample_id_column] = self.sample_metadata[
                    self._sample_id_column
                ].astype(str)
                self.sample_metadata.set_index(self._sample_id_column, inplace=True)
                self.data = parsed_data
                self.data.columns = [str(i) for i in self.data.columns]
                self.data.set_index(self._sample_id_column, inplace=True)
                # Drop samples not in data
                self.sample_metadata = self.sample_metadata.loc[self.data.index]
                self.samples = list(self.sample_metadata.index)
            else:
                raise ValueError("No sample ID column found in data")

            # check that metabolite ID column is found in chemical annotation
            if self._metabolite_id_column in parsed_chemical_annotation.columns:
                self.chemical_annotation = parsed_chemical_annotation
                self.chemical_annotation[self._metabolite_id_column] = (
                    self.chemical_annotation[self._metabolite_id_column].astype(str)
                )
                self.chemical_annotation.set_index(
                    self._metabolite_id_column, inplace=True
                )
                # Cleanup data from metadata
                self._remove_metadata_from_data()
                # Update the chemical annotation data in the class
                self._update_chemical_annotation()

            else:
                raise ValueError("No metabolite ID column found in chemical annotation")
        except ValueError as ve:
            raise ValueError(f"Error setting up data: {ve}")

    def import_excel(
        self,
        file_path: str,
        sample_metadata: str = "sample_metadata",
        chemical_annotation: str = "chemical_annotation",
        data_sheet: str = "data",
    ) -> None:
        """
        Import data from an Excel file and set up the class.

        Parameters
        ----------
        file_path : str
            Path to the Excel file.
        data_sheet : str
            Name of the sheet containing the main data.

        Raises
        ------
        FileNotFoundError
            If the Excel file is not found.
        AttributeError
            If the data sheet is not found in the Excel file.
        ValueError
            If there is an error importing the data.
        """
        try:
            self.parser.import_excel(
                file_path, sample_metadata, chemical_annotation, data_sheet
            )
            self._setup_data(
                self.parser.chemical_annotation,
                self.parser.sample_metadata,
                self.parser.generic_data,
            )
            self._remove_metadata_from_data()
        except FileNotFoundError as fnfe:
            raise FileNotFoundError(f"File '{file_path}' not found.") from fnfe
        except AttributeError as ae:
            raise AttributeError(
                f"Data sheet '{data_sheet}' not found in the Excel file."
            ) from ae
        except ValueError as ve:
            raise ValueError(f"Error importing data: {ve}") from ve

    def import_tables(
        self,
        data: Union[pd.DataFrame, str],
        chemical_annotation: Union[pd.DataFrame, str] = "config/metabolites.tsv",
        sample_metadata: Union[pd.DataFrame, str] = "config/samples.tsv",
    ) -> None:
        """
        Import data from tables and set up the class.

        Parameters:
            data: DataFrame or path of file containing the main data.
            chemical_annotation: DataFrame or path to the file containing chemical annotation.
            sample_metadata: DataFrame or path to the file containing sample metadata.

        """
        try:
            self.parser.import_tables(
                sample_metadata=sample_metadata,
                chemical_annotation=chemical_annotation,
                generic_data=data,
            )
            self._setup_data(
                self.parser.chemical_annotation,
                self.parser.sample_metadata,
                self.parser.generic_data,
            )
            self._remove_metadata_from_data()
        except ValueError as ve:
            raise ValueError("Error importing data: {}".format(ve)) from ve

    def merge_sample_metadata_data(self) -> pd.DataFrame:
        """
        Merge sample metadata and metabolite abundance data into a single DataFrame.

        This method merges the sample metadata and the metabolite abundance data
        by matching the sample IDs in both dataframes. The resulting DataFrame
        contains all the columns from both the sample metadata and the data.

        The sample IDs in the data are used as the index for the merge, and the
        resulting dataframe will have the same index as the data.

        Returns:
            DataFrame: Merged DataFrame containing sample metadata and data.

        """
        # Merge sample metadata and data by matching sample IDs
        merged = self.sample_metadata.merge(
            self.data, left_index=True, right_index=True, how="inner"
        )
        return merged

    def save_merged(self, data_path, chemical_annotation_path=None):
        """
        Save the merged sample data and metabolite abundance data to TSV files.

        This function merges the sample metadata and metabolite abundance data
        and saves the resulting DataFrame to a TSV file. The chemical
        annotation is saved optionally.

        Parameters:
        data_path (str): Path to save the merged sample data.
        chemical_annotation_path (str, optional): Path to save the chemical annotation data. Default is None.

        Raises:
        ValueError: If attempting to save an empty DataFrame.
        """
        merged = self.merge_sample_metadata_data()
        if len(merged) == 0:
            raise ValueError("Trying to save an empty dataframe")
        merged.to_csv(data_path, sep="\t")
        if chemical_annotation_path:
            self.chemical_annotation.to_csv(chemical_annotation_path, sep="\t")

    def save_tables(
        self,
        data_path: Optional[str] = None,
        chemical_annotation_path: Optional[str] = None,
        sample_metadata_path: Optional[str] = None,
    ) -> None:
        """
        Save individual parts of the dataset to TSV files.

        This function saves the individual parts of the dataset to TSV files. The
        parts that can be saved are:

        * `data`: The metabolite quantification data.
        * `chemical_annotation`: The chemical annotation data.
        * `sample_metadata`: The sample metadata.

        All parameters are optional, and the function will only save the parts
        for which a path was provided.

        Parameters:
            data_path (str, optional): Path to save the data. Default is None.
            chemical_annotation_path (str, optional): Path to save the chemical
                annotation data. Default is None.
            sample_metadata_path (str, optional): Path to save the sample metadata.
                Default is None.

        Raises:
            ValueError: If attempting to save an empty DataFrame.
        """
        if len(self.data) == 0:
            raise ValueError("Trying to save an empty dataframe")
        if data_path:
            self.data.to_csv(data_path, sep="\t")
        if chemical_annotation_path:
            self.chemical_annotation.to_csv(chemical_annotation_path, sep="\t")
        if sample_metadata_path:
            self.sample_metadata.to_csv(sample_metadata_path, sep="\t")

    def save_excel(self, file_path, data_sheet="data"):
        """
        Save the dataset to an Excel file.

        This function saves the dataset to an Excel file. The data, chemical
        annotation, and sample metadata are saved to separate sheets in the
        Excel file. The name of the sheet containing the data is specified by
        the `data_name` parameter.

        Parameters:
            file_path (str): Path to save the Excel file.
            data_name (str): Name of the sheet containing the data. Default is
                "data".

        Returns:
            None
        """
        with pd.ExcelWriter(file_path) as writer:
            self.chemical_annotation.to_excel(writer, sheet_name="chemical_annotation")
            self.sample_metadata.to_excel(writer, sheet_name="sample_metadata")
            self.data.to_excel(writer, sheet_name=data_sheet)

    def add_to_excel(self, file_path, new_sheet, new_sheet_name):
        """
        Add a new sheet to an Excel file.
        This function adds a new sheet to an Excel file.
        The new sheet is specified by the `new_sheet` parameter, and the name of
        the new sheet is specified by the `new_sheet_name` parameter.

        Parameters:
            file_path (str): Path of the Excel file to add the new sheet to.
            new_sheet (DataFrame): DataFrame to add as a new sheet.
            new_sheet_name (str): Name of the new sheet.

        Returns:
            None
        """
        with pd.ExcelWriter(file_path, mode="a") as writer:
            new_sheet.to_excel(writer, sheet_name=new_sheet_name)

    ###
    # Utility functions to manipulate the class attributes
    ###

    def replace_column_names(self, new_column: str) -> None:
        """
        Replace data column names with names from a column of the metabolite metadata.

        Replaces the names of the columns in the data with the names from the specified
        column of the metabolite metadata. The metabolite metadata is expected to
        have a column with the specified name. Additionally, this function checks
        that the chemical annotation is complete for all features in the dataset.

        Parameters:
            new_column (str): Name of the column from metabolite metadata to use for renaming.

        Raises:
            ValueError: If the specified column does not exist in the metabolite metadata or if
                chemical annotation is missing for some features.
        """
        # Check that the column exists in the metabolite metadata
        if new_column not in self.chemical_annotation.columns:
            raise ValueError(f"No column named {new_column} in the metabolite metadata")

        # Check that all features have annotation
        not_annotated = list(
            set(self.data.columns).difference(set(self.chemical_annotation.index))
        )
        if len(not_annotated) > 0:
            raise ValueError(f"Missing chemical annotation for {not_annotated}")

        # Create dictionary for renaming
        renaming_dict = {
            old: new
            for old, new in zip(
                self.chemical_annotation.index, self.chemical_annotation[new_column]
            )
        }

        # Perform the renaming
        self.data.columns = [renaming_dict[old] for old in self.data.columns]

        # Update the name of the column used for feature identification in the data
        self._metabolite_id_column = new_column

        # Reset the index of the chemical annotation to match the new column name and update
        self.chemical_annotation = self.chemical_annotation.reset_index().set_index(
            new_column
        )

        # Update the chemical annotation data in the class
        self._update_chemical_annotation()

    def _update_chemical_annotation(self):
        """
        Update the chemical annotation based on the current data columns.

        This function ensures that the chemical annotation DataFrame only contains
        the columns present in the data, and update the metabolites attribute with the
        new list of metabolites.
        """
        self.chemical_annotation = self.chemical_annotation.loc[list(self.data.columns)]
        self.metabolites = list(self.chemical_annotation.index)

    def _update_sample_metadata(self):
        """
        This function updates the sample metadata based on the current data index.

        It ensures that the sample metadata DataFrame only contains
        the rows present in the data, and updates the `samples` attribute with the
        new list of samples.
        """
        # Update the sample metadata DataFrame to only contain the rows present in the
        # data
        self.sample_metadata = self.sample_metadata.loc[self.data.index]

        # Update the samples attribute with the new list of samples
        self.samples = list(self.sample_metadata.index)

    def _remove_metadata_from_data(self):
        """
        Remove metadata columns from the data.

        This function removes the metadata columns from the data by using the columns
        present in the chemical annotation DataFrame.

        Raises:
            ValueError: If there is an error removing the metadata from the data.
        """
        metabolites_in_data = [
            i for i in self.chemical_annotation.index if i in self.data.columns
        ]
        try:
            self.data = self.data[metabolites_in_data]
        except ValueError as ve:
            raise ValueError("Error removing metadata from data: {}".format(ve))

    def _split_by_sample_column(self, column: list) -> Dict[str, "DatasetManager"]:
        """
        Split the dataset (data and sample metadata) in multiple independent DataClass instances
        based on the values of one or more sample metadata columns.

        This function splits the dataset into multiple independent DataClass instances, each
        containing a subset of the data based on the values of a sample metadata column. The
        function returns a dictionary containing the split data, where the dictionary keys are
        the unique values of the sample metadata column and the values are the DataClass instances
        containing the split data.

        Args:
            column: The name of the column in the sample metadata DataFrame to use for splitting.

        Returns:
            A dictionary containing the split data, where the dictionary keys are the unique
            values of the sample metadata column and the values are the DataClass instances
            containing the split data.
        """
        split_data: Dict[str, DatasetManager] = {}
        for name, group in self.sample_metadata.groupby(by=column):
            tempdata = self.data.loc[group.index]
            tempclass = DatasetManager(
                data_provider=self._data_provider,
                sample_id_column=self._sample_id_column,
                metabolite_id_column=self._metabolite_id_column,
            )
            tempclass.import_tables(
                data=tempdata.reset_index(),
                chemical_annotation=self.chemical_annotation.reset_index(),
                sample_metadata=group.reset_index(),
            )
            tempclass._update_chemical_annotation()
            tempclass._update_sample_metadata()
            tempclass._remove_metadata_from_data()
            split_data[name] = tempclass
        return split_data

    def _split_by_metabolite_column(self, column: str) -> Dict[str, "DatasetManager"]:
        """
        Split the data in multiple independent DataClass instances
        based on the values of a metabolite metadata column

        Parameters:
        column (str): Name of the column in the metabolite metadata table to use for splitting

        Returns:
        Dict[str, DatasetManager]: dictionary where keys are the unique values present in the column and values are the corresponding DataClass instances
        """
        split_data: Dict[str, "DatasetManager"] = {}
        for name, group in self.chemical_annotation.groupby(by=column):
            tempdata = self.data[list(group.index)]
            tempclass = DatasetManager(
                data_provider=self._data_provider,
                sample_id_column=self._sample_id_column,
                metabolite_id_column=self._metabolite_id_column,
            )
            tempclass.import_tables(
                data=tempdata.reset_index(),
                chemical_annotation=group.reset_index(),
                sample_metadata=self.sample_metadata.reset_index(),
            )
            tempclass._update_chemical_annotation()
            split_data[name] = tempclass
        return split_data

    def drop_samples(self, samples_to_drop, inplace=True):
        """
        Drop specified samples from the dataset.

        Parameters
        ----------
        samples_to_drop : list
            List of sample IDs to drop.
        inplace : bool, optional
            If True, drop the samples from the existing data and metadata tables, by default True

        Returns
        -------
        pandas.DataFrame or None
            If `inplace` is False, the resulting data after dropping the samples, otherwise None
        """
        if not isinstance(samples_to_drop, list):
            samples_to_drop = [samples_to_drop]
        remaining = self.data.drop(index=[str(i) for i in samples_to_drop])
        if inplace:
            self.data = remaining
            self._update_sample_metadata()
            return None
        else:
            return remaining

    def drop_metabolites(self, metabolites_to_drop, inplace=True):
        """
        Drop specified metabolites from the dataset.

        Parameters
        ----------
        metabolites_to_drop : list
            List of metabolite IDs to drop.
        inplace : bool, optional
            If True, drop the metabolites from the existing data and chemical annotation tables, by default True

        Returns
        -------
        pandas.DataFrame or None
            If `inplace` is False, the resulting data after dropping the metabolites, otherwise None
        """
        if not isinstance(metabolites_to_drop, list):
            metabolites_to_drop = [metabolites_to_drop]
        remaining = self.data.drop(columns=[str(i) for i in metabolites_to_drop])
        if inplace:
            self.data = remaining
            self._update_chemical_annotation()
            return None
        else:
            return remaining

    def drop_xenobiotic_metabolites(self, inplace=True):
        """
        Drop metabolites of xenobiotic origin from the dataset.

        Xenobiotics are metabolites that are not found in the human body and are
        considered as foreign substances.

        Parameters
        ----------
        inplace : bool, optional
            If True, drop the metabolites from the existing data and chemical annotation tables, by default True

        Returns
        -------
        pandas.DataFrame or None
            If `inplace` is False, the resulting data after dropping the metabolites, otherwise None
        """
        xenobiotic_metabolites = list(
            self.chemical_annotation[
                self.chemical_annotation["SUPER_PATHWAY"].str.lower() == "xenobiotics"
            ].index
        )
        if inplace:
            self.drop_metabolites(xenobiotic_metabolites, inplace=True)
            return None
        else:
            return self.drop_metabolites(xenobiotic_metabolites, inplace=False)

    def extract_metabolites(self, metabolites_to_extract):
        """
        Extract data for specified metabolites.

        This function extracts the abundance data for the specified metabolites
        and merges the resulting data with the sample metadata.

        Parameters
        ----------
        metabolites_to_extract : str or list
            Name(s) of metabolite(s) to extract.

        Returns
        -------
        pandas.DataFrame
            Extracted data for the specified metabolites merged with sample metadata
        """
        if not isinstance(metabolites_to_extract, list):
            metabolites_to_extract = [metabolites_to_extract]
        return self.sample_metadata.merge(
            self.data[[str(i) for i in metabolites_to_extract]],
            left_index=True,
            right_index=True,
        )

    def extract_samples(self, samples_to_extract):
        """
        Extract data for specified samples.

        This function extracts the sample metadata and abundance data
        for the specified samples.

        Parameters
        ----------
        samples_to_extract : str or list
            ID(s) of sample(s) to extract.

        Returns
        -------
        pandas.DataFrame
            Extracted sample metadata and abundance data for the specified samples.
        """
        if not isinstance(samples_to_extract, list):
            samples_to_extract = [samples_to_extract]
        return self.sample_metadata.loc[samples_to_extract].merge(
            self.data,
            left_index=True,
            right_index=True,
            how="left",
        )

    def extract_chemical_annotations(self, metabolites_to_extract):
        """
        Extract chemical annotations for the specified metabolites.

        This method extracts the chemical annotations for the specified metabolites.

        Parameters
        ----------
        metabolites_to_extract : str or list
            Name(s) of metabolite(s) to extract annotations for.

        Returns
        -------
        pandas.DataFrame
            Extracted chemical annotations for the specified metabolites.
        """
        if not isinstance(metabolites_to_extract, list):
            metabolites_to_extract = [metabolites_to_extract]
        # convert to string to match the CHEM_ID column in chemical_annotation
        return self.chemical_annotation.loc[[str(i) for i in metabolites_to_extract]]

    def sort_samples(self, by, ascending=True):
        """
        Sort samples in the dataset.

        Parameters
        ----------
        by : str
            Column name to sort by.
        ascending : bool, optional
            If True, sort in ascending order, by default True

        Returns
        -------
        pandas.DataFrame
            Sorted dataset
        """
        self.sample_metadata = self.sample_metadata.sort_values(
            by=by, ascending=ascending
        )
        self.data = self.data.loc[self.sample_metadata.index]
        self.samples = self.sample_metadata.index



import numpy as np

import pandas as pd
from metabotk.utils import validate_dataframe


class MissingDataHandler:
    """
    Class for detecting and counting missing values in data,
    as well as removing columns/rows based on their missing data content.

    Methods:
        detect_missing(data): Detects missing values in a collection.
        count_missing(data_frame, axis=0): Counts missing values in each row or column of a DataFrame.
        drop_columns_with_missing(data_frame): Removes columns with missing values above the threshold.
        drop_rows_with_missing(data_frame): Removes rows with missing values above the threshold.
    """

    def __init__(self):
        """
        Initializes the MissingDataHandler.
        """

    def _validate_threshold(self, threshold):
        """
        Validates the threshold attribute.
        """
        if not isinstance(threshold, (int, float)):
            raise TypeError("Threshold must be a numeric value")
        if not 0 <= threshold <= 1:
            raise ValueError("Threshold must be between 0 and 1")

    def _detect_missing(self, data) -> np.array:
        """
        Detects missing values in a collection.

        Parameters:
            data (list, array, or Series): Collection containing data.

        Returns:
            Boolean array indicating missing (True) and non-missing data (False).
        """

        is_missing = np.isnan(data)
        return is_missing

    def _count_missing(self, data):
        """
        Counts missing values in a collection.

        Parameters:
            data (list, array, or Series): Collection containing data.

        Returns:
            Number of missing values in collection
        """

        missing = self._detect_missing(data)
        n_missing = missing.sum()
        return n_missing

    def _count_missing_in_dataframe(self, data_frame, axis=0):
        """
        Counts missing values in each row or column of a DataFrame.

        Parameters:
            data_frame (DataFrame): Pandas DataFrame containing only numeric values.
            axis (int, optional): Axis along which to count missing values.
                0 for columns, 1 for rows. Default is 0.

        Returns:
            Series: Pandas Series with the row/column index and the number of missing values.
        """
        validate_dataframe(data_frame)
        missing_values = data_frame.apply(self._detect_missing, axis=axis)
        n_missing_values = missing_values.sum(axis=axis)
        return n_missing_values

    def _drop_columns_with_missing(self, data_frame, threshold=0.25):
        """
        Removes columns with missing values above the threshold.

        Parameters:
            data_frame (DataFrame): Pandas DataFrame containing only numeric values.

        Returns:
            DataFrame: DataFrame with columns with missingness higher than the threshold.
        """
        self._validate_threshold(threshold)
        validate_dataframe(data_frame)
        missing = self._count_missing_in_dataframe(data_frame, axis=0)
        to_drop = missing[missing / len(data_frame) > threshold]
        to_drop = data_frame[list(to_drop.index)]
        return to_drop

    def _drop_rows_with_missing(self, data_frame, threshold=0.25):
        """
        Removes rows with missing values above the threshold.

        Parameters:
            data_frame (DataFrame): Pandas DataFrame containing only numeric values.

        Returns:
            DataFrame: DataFrame with rows with missingness higher than the threshold.
        """
        self._validate_threshold(threshold)
        validate_dataframe(data_frame)
        missing = self._count_missing_in_dataframe(data_frame, axis=1)
        to_drop = missing[missing / len(data_frame.columns) > threshold]
        to_drop = data_frame.loc[to_drop.index]
        return to_drop

    def _drop_missing_from_dataframe(
        self, data_frame: pd.DataFrame, axis: int = 0, threshold: float = 0.25
    ) -> pd.DataFrame:
        """
        Remove columns or rows with missing values above the threshold.

        Parameters:
            - threshold: missingness over which to remove the row/column
            - axis: {0 or ‘index’, remove columns, 1 or ‘columns’, remove rows}, default 0

        Returns:
            DataFrame: DataFrame containing the rows/columns dropped.
        """
        if axis == 0:
            to_drop = self._drop_columns_with_missing(
                data_frame=data_frame, threshold=threshold
            )
            print(f"Removed {len(to_drop.columns)} columns")
            remaining_data = data_frame.drop(columns=to_drop.columns)
            return remaining_data

        if axis == 1:
            to_drop = self._drop_rows_with_missing(
                data_frame=data_frame, threshold=threshold
            )
            print(f"Removed {len(to_drop.index)} rows")
            remaining_data = data_frame.drop(index=to_drop.index)
            return remaining_data



"""
Visualization module for MetaboTK
This module provides functions to visualize the results of principal component analysis (PCA) and plot metabolite
abundance data.
TODO: add more visualizations
"""

import seaborn as sns
import pandas as pd
from typing import Optional
from metabotk.dimensionality_reduction import DimensionalityReduction


class Visualization:
    def __init__(self, data_manager) -> None:
        """
        Initialize the class.

        Parameters:
        ----------
        data_manager: Object
            Instance of DataManager class
        """
        self.data_manager = data_manager
        """Instance of DataManager class"""

        self.dimensionality_reduction = DimensionalityReduction(
            data_manager=self.data_manager
        )
        """Instance of DimensionalityReduction class"""

    def plot_pca(self, pca=None, x="PC1", y="PC2", hue=None, style=None, savepath=None):
        """
        Plot the results of principal component analysis (PCA).

        This function plots the PCA results using seaborn's scatterplot function.
        If no PCA data is provided, it will compute the PCA with 3 components
        and plot the results.

        Parameters
        ----------
        pca: pandas.DataFrame, optional
            DataFrame with PCA results. If not provided, it will be computed
            using the `get_pca` function from the `DimensionalityReduction` class.
        x: str, default='PC1'
            Name of the X-axis.
        y: str, default='PC2'
            Name of the Y-axis.
        hue: str, optional
            Column name in sample metadata DataFrame to color the points by.
        style: str, optional
            Column name in sample metadata DataFrame to shape the points by.
        savepath: str, optional
            Path to save the plot as an image file. If not provided, the plot
            will not be saved.

        Returns
        -------
        plot: seaborn.axisgrid.FacetGrid
            Seaborn scatterplot object.
        """
        if not pca:
            print("PCA not found, computing now with 3 components...")
            pca = self.dimensionality_reduction.get_pca(n_components=3)
        plot = sns.scatterplot(data=pca, x=x, y=y, hue=hue, style=style)
        if savepath:
            plot.figure.savefig(savepath)
        return plot

    def plot_pca_grid(
        self,
        pca: pd.DataFrame = None,
        hue: Optional[str] = None,
        savepath: Optional[str] = None,
    ) -> sns.axisgrid.PairGrid:
        """
        Plot PCA results in a grid using seaborn's pairplot.

        Parameters
        ----------
        pca: pandas.DataFrame, optional
            DataFrame with PCA results. If not provided, it will be computed
            using the `get_pca` function from the `DimensionalityReduction` class.
        hue: str, optional
            Column name in sample metadata DataFrame to color the points by.
        savepath: str, optional
            Path to save the plot as an image file. If not provided, the plot
            will not be saved.

        Returns
        -------
        plot: seaborn.axisgrid.PairGrid
            Seaborn pairplot object.
        """
        if pca is None:
            print("PCA not found, computing now with 3 components...")
            pca = self.dimensionality_reduction.get_pca(n_components=3)
        plot = sns.pairplot(
            data=pca,
            vars=[col for col in pca.columns if col.startswith("PC")],
            hue=hue,
            diag_kind="kde",
            diag_kws={"linewidth": 0, "shade": False},
        )

        if savepath:
            plot.fig.savefig(savepath)

    def plot_metabolite(self, metabolite, x=None, hue=None, savepath=None):
        """
        Plot metabolite abundance data.

        Parameters
        ----------
        metabolite : str
            Name of the metabolite.
        x : str, optional
            Name of the X-axis. Default is the class sample ID column.
        hue : str, optional
            Column name in sample metadata DataFrame to color the points by.
            Default is None.
        savepath : str, optional
            Path to save the plot as an image file. Default is None.

        Returns
        -------
        plot : seaborn.axisgrid.FacetGrid
            Seaborn scatterplot object.
        """
        if not x:
            x = self.data_manager._sample_id_column
        data = self.data_manager.extract_metabolites(metabolite)
        plot = sns.scatterplot(data=data, x=x, y=str(metabolite), hue=hue)
        if savepath:
            plot.figure.savefig(savepath)
        return plot



import pandas as pd
from sklearn.preprocessing import StandardScaler
from sklearn.decomposition import PCA
import sys


class DimensionalityReduction:
    """
    Class for performing dimensionality reduction on metabolomics data.

    This class provides a simple interface to perform Principal Component
    Analysis (PCA) on the data.
    """

    def __init__(self, data_manager):
        """
        Initialize the class.

        Parameters:
            data_manager (DataManager): DataManager instance containing the data.
        """
        self.data_manager = data_manager

    def get_pca(self, n_components=3, get_pca_object=False):
        """
        Perform Principal Component Analysis (PCA) on the data.

        Parameters:
            n_components (int, optional): Number of components for the PCA. Default is 3.

        Raises:
            ValueError: If the data contains NaN values or not all columns contain numeric data.

        Returns:
            pca_transformed (DataFrame): DataFrame containing the PCA-transformed data.
            pca (PCA): sklearn PCA object.
        """
        input_data = self.data_manager.data
        # Check if all columns have numeric data types
        if input_data.isnull().any().any():
            # Ignore columns with empty values
            print(
                "Data contains NaN values; columns with empty values were ignored",
                file=sys.stderr,
            )
            input_data = input_data.dropna(axis=1)
        if not input_data.select_dtypes(include=["number"]).equals(input_data):
            raise ValueError("Not all columns contain numeric data.")

        # Normalize data
        scaler = StandardScaler()
        scaled_data = scaler.fit_transform(input_data)

        # Perform PCA
        pca = PCA(n_components=n_components).fit(scaled_data)

        # Transform data
        pca_transformed = pd.DataFrame(
            pca.transform(scaled_data),
            columns=[f"PC{i}" for i in range(1, n_components + 1)],
            index=self.data_manager.data.index,
        )

        # Concatenate with sample metadata
        pca_transformed = self.data_manager.sample_metadata.merge(
            pca_transformed, left_index=True, right_index=True
        )
        if get_pca_object == True:
            return pca_transformed, pca
        else:
            return pca_transformed



from typing import List
import pandas as pd
import pyserrf as srf
from skloess import LOESS


class NormalizationHandler:
    """
    Class for performing normalization on metabolomics data.

    This class provides a simple interface to perform normalization on the data using various approaches.
    For now only SERRF is included.
    """

    def __init__(
        self,
        dataset_manager,
    ):
        """
        Initialize the class.
        """
        self._dataset_manager = dataset_manager

    def serrf(
        self,
        dataset,
        sample_type_column="measurement_group",
        batch_column="batch",
        time_column="time",
        other_columns=None,
        n_correlated_metabolites=10,
        random_state=None,
        threads=1,
        cross_validation=False,
    ):
        """
        Perform normalization using SERRF.
        Parameters
        ----------
        sample_type_column : str, optional
            The name of the column in the sample metadata with the sample type
            information (i.e qc or normal sample). The default value is
            'sampleType'.
        batch_column : str, optional
            The name of the column in the sample metadata with the batch
            information. If None, all samples are considered as part the same
            batch. The default value is 'batch'.
        time_column: str, optional
            The name of the column in the sample metadata with the injection time
            information.The default value is 'time'.
        other_columns : list of str or None, optional
            A list with the names of other metadata columns in the dataset; it is
            important to specify all the metadata columns to separate them from
            the metabolite abundance values. The default value is None
        random_state : int, RandomState instance, or None, optional
            The random seed used for all methods with a random component (i.e
            numpy normal distribution, sklearn random forest regressor). The
            default value is None, which means that a random seed is generated
            automatically. To obtain reproducible results, set a specific random
            seed.
        n_correlated_metabolites : int, optional
            The number of metabolites with the highest correlation to the
            metabolite to be normalized. The default value is 10.
        threads: int, optional
            Number of threads to use for parallel processing        Returns:
                dict: Dictionary with imputed datasets, keys are integers starting from 1.
        """
        serrf_instance = srf.SERRF(
            sample_type_column=sample_type_column,
            batch_column=batch_column,
            time_column=time_column,
            other_columns=other_columns,
            n_correlated_metabolites=n_correlated_metabolites,
            random_state=random_state,
            threads=threads,
        )
        normalized = serrf_instance.fit_transform(dataset, return_data_only=True)
        return normalized

    def serrf_CV(
        self,
        dataset,
        n_splits=5,
        sample_type_column="measurement_group",
        batch_column="batch",
        time_column="time",
        other_columns=None,
        n_correlated_metabolites=10,
        random_state=None,
        threads=1,
    ):
        """
        Perform cross-validation using SERRF.

        Parameters
        ----------
        dataset : DataFrame
            The dataset to perform cross-validation on.
        n_splits : int, optional
            The number of splits to perform (default: 5).
        sample_type_column : str, optional
            The name of the column in the sample metadata with the sample type
            information (i.e qc or normal sample). The default value is
            'sampleType'.
        batch_column : str, optional
            The name of the column in the sample metadata with the batch
            information. If None, all samples are considered as part the same
            batch. The default value is 'batch'.
        time_column: str, optional
            The name of the column in the sample metadata with the injection time
            information.The default value is 'time'.
        other_columns : list of str or None, optional
            A list with the names of other metadata columns in the dataset; it is
            important to specify all the metadata columns to separate them from
            the metabolite abundance values. The default value is None
        n_correlated_metabolites : int, optional
            The number of metabolites with the highest correlation to the
            metabolite to be normalized. The default value is 10.
        random_state : int, RandomState instance, or None, optional
            The random seed used for all methods with a random component (i.e
            numpy normal distribution, sklearn random forest regressor). The
            default value is None, which means that a random seed is generated
            automatically. To obtain reproducible results, set a specific random
            seed.
        threads: int, optional
            Number of threads to use for parallel processing.

        Returns
        -------
        dict
            Dictionary with cross-validated datasets, keys are integers starting from 1.
        """
        # Perform cross-validation
        raw_variations, normalized_variations = srf.cross_validate(
            dataset,
            n_splits=n_splits,
            sample_type_column=sample_type_column,
            batch_column=batch_column,
            time_column=time_column,
            other_columns=other_columns,
            n_correlated_metabolites=n_correlated_metabolites,
            random_state=random_state,
            threads=threads,
        )
        return raw_variations, normalized_variations

    def loess_single_metabolite(
        self,
        metabolite: str,
        degree: int,
        smoothing: float,
        rescale: bool,
        qc_samples: List[str],
        injection_order_col: str,
    ) -> pd.Series:
        """
        Perform LOESS normalization on a single metabolite.

        Parameters
        ----------
        metabolite : str
            The name of the metabolite to normalize.
        degree : int
            The degree of the LOESS polynomial.
        smoothing : float
            The smoothing parameter for LOESS.
        rescale : bool
            Whether to rescale the normalized values to the median.
        qc_samples : List[str]
            The names of the QC samples.
        injection_order_col : str
            The name of the column containing the injection order.

        Returns
        -------
        pd.Series
            The normalized values for the metabolite.
        """
        # Extract the metabolite and remove missing values
        data = self._dataset_manager.extract_metabolites(metabolite)

        # Extract the QC samples
        qc = data.loc[qc_samples]
        x_qc = qc[injection_order_col]
        y_qc = qc[metabolite]

        # Perform LOESS normalization
        loess = LOESS(degree=degree, smoothing=smoothing)
        loess.fit(x_qc.values, y_qc.values)
        pred = loess.predict(data[injection_order_col].values)
        norm = data[metabolite] - pred

        # Rescale the normalized values if requested
        if rescale:
            median = data[metabolite].median()
            norm = norm + median

        return norm

    def loess(
        self,
        degree: int,
        smoothing: float,
        rescale: bool,
        qc_samples: List[str],
        injection_order_col: str,
        inplace: bool = True,
    ):
        """
        Perform LOESS normalization on all metabolites in the dataset.

        Parameters
        ----------
        degree : int
            The degree of the LOESS polynomial.
        smoothing : float
            The smoothing parameter for LOESS.
        rescale : bool
            Whether to rescale the normalized values to the median.
        qc_samples : List[str]
            The names of the QC samples.
        injection_order_col : str
            The name of the column containing the injection order.
        inplace : bool, optional
            Whether to perform the normalization inplace. Default is True.

        Returns
        -------
        pd.DataFrame
            The normalized data.
        """
        # Loop over all metabolites in the dataset
        data = self._dataset_manager.data
        for metabolite in self._dataset_manager.metabolites:
            normalized_metabolite = self.loess_single_metabolite(
                metabolite,
                degree,
                smoothing,
                rescale,
                qc_samples,
                injection_order_col,
            )
            data[metabolite] = normalized_metabolite
        if inplace:
            self._dataset_manager.data = data
        else:
            return data



"""
Statistics module
"""

import pandas as pd
import numpy as np
from metabotk.outliers_handler import OutlierHandler
from metabotk.missing_handler import MissingDataHandler
from metabotk.utils import ensure_numeric_data


def compute_correlations(data_frame: pd.DataFrame, method: str) -> pd.DataFrame:
    """
    Computes correlations between columns in a pandas DataFrame.

    Parameters:
        data_frame (DataFrame): Pandas DataFrame containing numerical values.
        method (str): Method for computing correlations. Default is 'pearson'.

    Returns:
        DataFrame: Pandas DataFrame containing correlations between columns.
    """
    correlations = data_frame.corr(method=method)
    return correlations


def get_top_n_correlations(data_frame: pd.DataFrame, n=10, method="pearson"):
    """
    Get the top n correlations for each column in a pandas DataFrame.

    Parameters:
        data_frame (DataFrame): Pandas DataFrame containing numerical values.
        n (int): Number of top correlations to return. Default is 10.
        method (str): Method for computing correlations. Default is 'pearson'.

    Returns:
        DataFrame: Pandas DataFrame containing top n correlations
        for each id and their values.
    """
    correlations = compute_correlations(data_frame, method)
    top_correlations = []

    for id in correlations.columns:
        top_n_abs = (
            correlations[id].abs().sort_values(ascending=False).drop(id).iloc[0:n].index
        )
        top_n = correlations[id].loc[top_n_abs]
        top_n.index.name = "correlated_ids"
        top_n.name = "correlation"
        top_n = top_n.reset_index()
        top_n["id"] = id
        top_correlations.append(top_n)
    top_correlations = pd.concat(top_correlations)
    top_correlations = top_correlations[["id", "correlated_ids", "correlation"]]
    return top_correlations


class StatisticsHandler:
    """
    Class for obtaining basic statistics about the data.

    This class provides methods for computing basic statistics
    (mean, standard deviation, median, min, max, sum, CV%, missing, outliers)
    for a collection of numerical data or a pandas DataFrame.
    """

    def __init__(self, dataset_manager):
        """
        Initializes the StatisticsHandler.
        """
        self._outlier_handler = OutlierHandler()
        self._missing_handler = MissingDataHandler()
        self._dataset_manager = dataset_manager

    def coefficient_of_variation(self, data):
        """
        Compute the coefficient of variation in percentage.

        The coefficient of variation (CV) is a measure of the dispersion of a dataset.
        It is calculated as the standard deviation (σ) divided by the mean (μ).
        The CV is expressed as a percentage (%).

        Parameters:
            data (list, array, or Series): Collection containing numerical data.

        Returns:
            float: Coefficient of Variation % (CV%).
        """
        data = ensure_numeric_data(data)
        # Remove any missing (nan) values.
        data = data[~np.isnan(data)]
        # Compute the standard deviation.
        std = np.std(data)
        # Compute the mean.
        mean = np.mean(data)
        # Compute the coefficient of variation.
        cv = std / mean
        # Convert coefficient of variation to a percentage.
        cv_pctg = cv * 100
        return cv_pctg

    def total_sum_abundance(self, data_frame, exclude_incomplete=True):
        """
        Computes total sum abundance (TSA) row-wise

        Parameters:
            data_frame (DataFrame): pandas DataFrame containing numerical data.
            exclude_incomplete (bool): option to exclude columns containing incomplete values
            from the computation

        Returns:
            Series: Pandas Series containing TSA values for each row.
        """
        if exclude_incomplete:
            data_frame = data_frame.dropna(axis=1).copy()
        else:
            data_frame = data_frame.copy()
        tsa = data_frame.sum(axis=1, skipna=True)
        tsa.name = "TSA"
        return tsa

    def compute_statistics(self, data, outlier_threshold):
        """
        Computes basic statistics for a collection of numerical data.

        Parameters:
            data (list, array, or Series): Collection containing numerical data.
            outlier_threshold (float): Threshold for outlier detection.

        Returns:
            Series: Pandas Series containing basic statistics.

        Notes:
            The statistics computed are:
            - Mean
            - Standard deviation
            - Median
            - Min
            - Max
            - Sum
            - Coefficient of Variation (CV%)
            - Number of missing values
            - Number of outliers
        """
        data_series = pd.Series(data)
        if len(data) == 0:
            raise ValueError("Input data is empty")
        stats = data_series.describe()
        cv = self.coefficient_of_variation(data)
        stats["CV%"] = cv
        stats = stats.rename(index={"50%": "median"})
        stats["missing"] = self._missing_handler._count_missing(data_series)
        stats["outliers"] = sum(
            self._outlier_handler.detect_outliers(
                data_series, threshold=outlier_threshold
            )
        )
        return stats

    def compute_dataframe_statistics(self, outlier_threshold, axis):
        """
        Computes basic statistics for a pandas DataFrame.

        Parameters:
            data_frame (DataFrame): Pandas DataFrame containing numerical values.
            outlier_threshold (float): Threshold for outlier detection. Default is None.
            axis (int): Which axis to compute statistics on. Default is 0 (column-wise)

        Returns:
            DataFrame: Pandas DataFrame containing statistics for each column.

        Notes:
            The statistics computed are:
            - Mean
            - Standard deviation
            - Median
            - Min
            - Max
            - Sum
            - Coefficient of Variation (CV%)
            - Number of missing values
            - Number of outliers
        """
        stats = self._dataset_manager.data.apply(
            lambda x: self.compute_statistics(x, outlier_threshold=outlier_threshold),
            axis=axis,
        )
        if axis == 0:
            stats = stats.transpose()
        return stats

    def metabolite_stats(self, outlier_threshold=5):
        """
        Computes basic statistics for the metabolomics data metabolite-wise

        This function computes basic statistics for each metabolite in the data.
        The statistics computed are:
        - Mean
        - Standard deviation
        - Median
        - Min
        - Max
        - Sum
        - Coefficient of Variation (CV%)
        - Number of missing values
        - Number of outliers

        Parameters:
        - outlier_threshold: Threshold for outlier detection (default=5).

        Returns:
            pandas DataFrame: DataFrame containing statistics for each metabolite.
            The index of the DataFrame is the metabolite names.

        """
        # Ensure that data is set up properly
        if self._dataset_manager.data.empty:
            raise ValueError(
                "No data available. Please import data before computing statistics."
            )

        # Compute statistics using StatisticsHandler
        metabolite_stats = self.compute_dataframe_statistics(outlier_threshold, axis=0)
        # self.metabolite_stats=metabolite_stats
        return metabolite_stats

    def sample_stats(self, outlier_threshold=5):
        """
        Computes basic statistics for the metabolomics data sample-wise

        This function computes statistics for each sample across all metabolites.

        Parameters:
            outlier_threshold (int): Threshold for identifying outliers.

        Returns:
             DataFrame containing statistics for each sample across all metabolites.
        """
        # Ensure that data is set up properly
        if self._dataset_manager.data.empty:
            raise ValueError(
                "No data available. Please import data before computing statistics."
            )
        sample_stats = self.compute_dataframe_statistics(outlier_threshold, axis=1)

        # Compute Total Sum of Abundance (TSA) for each sample across all metabolites
        tsa_complete_only = self.total_sum_abundance(
            self._dataset_manager.data, exclude_incomplete=True
        )
        tsa_complete_only.name = "TSA_complete_only"
        tsa_all = self.total_sum_abundance(
            self._dataset_manager.data, exclude_incomplete=False
        )
        tsa_all.name = "TSA_including_incomplete"
        tsa = pd.concat([tsa_complete_only, tsa_all], axis=1)

        # Merge TSA with sample statistics
        sample_stats = sample_stats.merge(tsa, left_index=True, right_index=True)
        # self.sample_stats=sample_stats
        return sample_stats



from metabotk.outliers_handler import OutlierHandler
from metabotk.missing_handler import MissingDataHandler
from metabotk.statistics_handler import StatisticsHandler
from metabotk.dataset_manager import DatasetManager
from metabotk.models_handler import ModelsHandler
from metabotk.visualization_handler import Visualization
from metabotk.dimensionality_reduction import DimensionalityReduction
from metabotk.imputation import ImputationHandler
from metabotk.feature_selection import FeatureSelection
from metabotk.normalization import NormalizationHandler
from typing import List, Dict


class MetaboTK(DatasetManager):
    """
    Class for working with metabolomics data
    """

    def __init__(
        self,
        data_provider="metabolon",
        sample_id_column=None,
        metabolite_id_column="CHEM_ID",
    ) -> None:
        """
        Initialize the class.
        """
        super().__init__(
            data_provider=data_provider,
            sample_id_column=sample_id_column,
            metabolite_id_column=metabolite_id_column,
        )

    @property
    def missing(self):
        """Lazy initialization of MissingDataHandler instance."""
        if not hasattr(self, "_missing_"):
            self._missing_ = MissingDataHandler()
        return self._missing_

    @property
    def outliers(self):
        """Lazy initialization of OutlierHandler instance."""
        if not hasattr(self, "_missing_"):
            self._outliers_ = OutlierHandler()
        return self._outliers_

    @property
    def stats(self):
        """Lazy initialization of StatisticsHandler instance."""
        if not hasattr(self, "_stats_"):
            self._stats_ = StatisticsHandler(self)
        return self._stats_

    @property
    def models(self):
        """Lazy initialization of ModelsHandler instance."""
        if not hasattr(self, "_models_"):
            self._models_ = ModelsHandler(self)
        return self._models_

    @property
    def dimensionality_reduction(self):
        """Lazy initialization of DimensionalityReduction instance."""
        if not hasattr(self, "_dimensionality_reduction_"):
            self._dimensionality_reduction_ = DimensionalityReduction(self)
        return self._dimensionality_reduction_

    @property
    def visualization(self):
        """Lazy initialization of Visualization instance."""
        if not hasattr(self, "_visualization_"):
            self._visualization_ = Visualization(self)
        return self._visualization_

    @property
    def imputation(self):
        """Lazy initialization of ImputationHandler instance."""
        if not hasattr(self, "_imputation_"):
            self._imputation_ = ImputationHandler(self)
        return self._imputation_

    @property
    def normalization(self):
        """Lazy initialization of NormalizationHandler instance."""
        if not hasattr(self, "_normalization_"):
            self._normalization_ = NormalizationHandler()
        return self._normalization_

    @property
    def feature_selection(self):
        """Lazy initialization of FeatureSelection instance."""
        if not hasattr(self, "_feature_selection_"):
            self._feature_selection_ = FeatureSelection(self)
        return self._feature_selection_

    ###FUNCTIONS###
    def drop_missing_from_dataframe(self, axis=0, threshold=0.25, inplace=False):
        """
        Removes rows/columns from the data dataframe based on the threshold of missing
        values.

        Parameters:
            axis (int): Axis to drop missing values from (0: rows, 1: columns).
            threshold (float): Threshold of missing values to drop.
            inplace (bool): Whether to drop missing values inplace or return the remaining data.

        Returns:
            DataFrame: DataFrame with missing values over threshold removed.
        """
        remaining_data = self.missing._drop_missing_from_dataframe(
            data_frame=self.data, axis=axis, threshold=threshold
        )
        if inplace:
            self.data = remaining_data
            self._update_chemical_annotation()
            print("Removed inplace metabolites with missing data over threshold")
            return None
        else:
            return remaining_data

    def split_by_sample_column(self, columns: List[str]) -> Dict[str, "MetaboTK"]:
        """
        Split the data based on the sample column(s).

        Parameters:
            columns (list): List of column names containing the sample IDs.

        Returns:
            dict: Dictionary with the split data and sample metadata, as a MetaboTK instance.
        """
        split_data = self._split_by_sample_column(columns)
        split_data_instanced = {}
        for dataset_name, dataset_instance in split_data.items():
            temp_instance = MetaboTK(
                sample_id_column=self._sample_id_column,
                data_provider=self._data_provider,
                metabolite_id_column=self._metabolite_id_column,
            )
            temp_instance.import_tables(
                data=dataset_instance.data.reset_index(),
                sample_metadata=dataset_instance.sample_metadata.reset_index(),
                chemical_annotation=dataset_instance.chemical_annotation.reset_index(),
            )
            split_data_instanced[dataset_name] = temp_instance
        return split_data_instanced

    def split_by_metabolite_column(self, columns: List[str]) -> Dict[str, "MetaboTK"]:
        """
        Split the data based on the metabolite column(s).

        Parameters:
            columns (list): List of column names containing the sample IDs.

        Returns:
            dict: Dictionary with the split data and sample metadata as a MetaboTK instance.
        """
        split_data = self._split_by_metabolite_column(columns)
        split_data_instanced = {}
        for dataset_name, dataset_instance in split_data.items():
            temp_instance = MetaboTK(
                sample_id_column=self._sample_id_column,
                data_provider=self._data_provider,
                metabolite_id_column=self._metabolite_id_column,
            )
            temp_instance.import_tables(
                data=dataset_instance.data.reset_index(),
                sample_metadata=dataset_instance.sample_metadata.reset_index(),
                chemical_annotation=dataset_instance.chemical_annotation.reset_index(),
            )
            split_data_instanced[dataset_name] = temp_instance
        return split_data_instanced



#!/usr/bin/env python
from typing import Union
import os
import argparse
from metabotk.interface import MetaboTK
from metabotk.utils import create_directory


def cli():
    """
    Command line interface for MetaboTK.

    This function is the entry point for the command line interface of MetaboTK. It sets up
    the command line arguments and passes them to the main analysis function.

    """

    parser = argparse.ArgumentParser(description="Metabolomics data analysis tool")

    parser.add_argument_group()

    ###INPUT ARGUMENTS

    input_group = parser.add_argument_group(
        "Input", "Options for reading in input data"
    )
    input_options = input_group.add_mutually_exclusive_group()
    input_options.add_argument(
        "-ie",
        "--excel",
        dest="excel",
        nargs=2,
        metavar=("file", "data_sheet"),
        help="Specify the excel file and the sheet from which to extract data",
    )
    input_options.add_argument(
        "-it",
        "--tables",
        dest="tables",
        nargs=3,
        metavar=("data", "samples", "metabolites"),
        help="Specify three tsv files: data, samples, and metabolites, in this order",
    )
    input_group.add_argument(
        "-sid",
        "--sample_id",
        dest="sample_id",
        metavar=("colname"),
        help="Specify the column name for sample IDs (must be present in both sample metadata and data); if not specified, the program will try to use the name of the first column in the data table",
    )
    input_group.add_argument(
        "-mid",
        "--metabolite_id",
        dest="metabolite_id",
        metavar=("colname"),
        help="Specify the column name for metabolite IDs in the chemical annotation; the metabolite IDs\
            in the column must correspond to the column names of the data table.\
            If not specified, the program will try to use CHEM_ID",
    )

    ###OUTPUT ARGUMENTS

    output_group = parser.add_argument_group("Output", "Output options")
    output_options = output_group.add_mutually_exclusive_group()
    output_options.add_argument(
        "-oe",
        "--output-excel",
        dest="output_excel",
        nargs=2,
        metavar=("file", "data_sheet"),
        help="Save the dataset to an excel file; if the file already exists, only the data will be appended as a new sheet",
    )
    output_options.add_argument(
        "-ot",
        "--output-tables",
        dest="output_tables",
        nargs=3,
        metavar=("data", "metabolites", "samples"),
        help="Save the dataset to three tsv files: data, metabolites and samples, in this order",
    )

    output_options.add_argument(
        "-od",
        "--output-data",
        dest="output_data",
        nargs=1,
        metavar=("data"),
        help="Save only the data to a TSV file",
    )

    ###SPLIT DATASET

    manipulation_group = parser.add_argument_group(
        "Manipulation", "Dataset manipulation"
    )
    manipulation_options = manipulation_group.add_mutually_exclusive_group()
    manipulation_options.add_argument(
        "-ss",
        "--split-samples",
        dest="split_samples",
        nargs=3,
        metavar=("sample_column", "output_dir", "output_format"),
        default=None,
        help="Split the dataset into multiple DataClass instances based on the values of a sample metadata column, and save each instance to a separate file in the output directory",
    )
    manipulation_options.add_argument(
        "-sm",
        "--split-metabolites",
        dest="split_metabolites",
        nargs=3,
        metavar=("metabolite_column", "output_dir", "output_format"),
        default=None,
        help="Split the dataset into multiple DataClass instances based on the values of a metabolite metadata column, and save each instance to a separate file in the output directory",
    )
    manipulation_options.add_argument(
        "--stratified-kfold",
        dest="stratified_kfold",
        nargs=3,
        metavar=("n_splits", "stratification_column", "output_dir"),
        default=None,
        help="Split the dataset into n_splits folds, and save each training and validation group to a separate file in the output directory. The samples in the folds will be stratified on the specified column",
    )

    ###ANALYSIS ARGUMENTS

    analysis_group = parser.add_argument_group("Analysis", "Analysis options")
    analysis_options = analysis_group.add_mutually_exclusive_group()
    analysis_group.add_argument(
        "--exclude-xenobiotics",
        dest="exclude_xenobiotics",
        action="store_true",
        help="Exclude xenobiotic metabolites from the analysis",
    )
    analysis_group.add_argument(
        "--outlier-threshold",
        dest="outlier_threshold",
        nargs="?",
        default=5,
        type=float,
        help="Numeric threshold for the outlier detection; default value is 5",
    )

    analysis_options.add_argument(
        "--sample-stats",
        action="store_true",
        dest="sample_stats",
        help="Print to stdout the sample statistics for the entire dataset",
    )

    analysis_options.add_argument(
        "--metabolite-stats",
        dest="metabolite_stats",
        action="store_true",
        help="Print to stdout each metabolite's statistics for the entire dataset",
    )
    analysis_options.add_argument(
        "--remove-outliers",
        dest="remove_outliers",
        action="store_true",
        help="Remove metabolite outlier values from the dataset and replace them with NAs; \
            outliers are computed column-wise;the data without outliers will be printed to stdout",
    )

    analysis_options.add_argument(
        "--fit-model",
        dest="fit_model",
        nargs="+",
        metavar=("formula", "model_dir"),
        help="Fit a linear model to each metabolite and print the residuals to stdout; optionally, save the models as pickle files in a folder of choice",
    )

    analysis_options.add_argument(
        "--PCA",
        dest="PCA",
        nargs="?",
        default=None,
        help="Perform a PCA on the dataset and print the results to stdout; number of components must be specified",
    )
    ###FEATURE SELECTION
    fs_group = parser.add_argument_group(
        "Feature Selection", "Feature selection methods and options"
    )
    fs_group.add_argument(
        "--feature-selection",
        dest="fs_method",
        nargs="?",
        default=None,
        help="Which feature selection method to use, default is boruta",
    )

    fs_group.add_argument(
        "-y",
        "--y_column",
        dest="y_column",
        help="Column containing target values",
        default=None,
    )
    fs_group.add_argument(
        "-t",
        "--threads",
        dest="threads",
        help="Threads to use, default is 1",
        default=1,
        type=int,
    )
    fs_group.add_argument(
        "-a",
        "--alpha",
        dest="alpha",
        help="Alpha (p-value threshold), default is 0.01",
        default=0.01,
        type=float,
    )

    fs_group.add_argument(
        "-r",
        "--random-state",
        dest="random_state",
        help="Random state seed, default is 42",
        default=42,
        type=int,
    )
    fs_group.add_argument(
        "-d",
        "--max-depth",
        dest="max_depth",
        help="BORUTA:Max depth of the tree, default is None",
        default=None,
        type=Union[None, int],
    )
    fs_group.add_argument(
        "-w",
        "--class-weight",
        dest="class_weight",
        help="BORUTA: Class weights, default is balanced",
        default="balanced",
    )
    fs_group.add_argument(
        "-n",
        "--n-estimators",
        dest="n_estimators",
        help="BORUTA: Number of estimators, default is 1000",
        default="auto",
    )
    fs_group.add_argument(
        "-i",
        "--max-iterations",
        dest="max_iterations",
        help="BORUTA: Max iterations. default is 1000",
        default=1000,
        type=int,
    )
    fs_group.add_argument(
        "-o",
        "--output-dir",
        dest="output_dir",
        help="BORUTA: Output directory for saving rankings, model and importance history, default is None (print ranking only to stdout)",
        default=None,
    )

    args = parser.parse_args()

    ########SET UP METABOTK INSTANCE

    if args.sample_id:
        sample_id_column = args.sample_id
    else:
        sample_id_column = None

    if args.metabolite_id:
        metabolite_id_column = args.metabolite_id
    else:
        metabolite_id_column = "CHEM_ID"

    metabotk_instance = MetaboTK(
        sample_id_column=sample_id_column, metabolite_id_column=metabolite_id_column
    )

    ###READ INPUT DATA
    if args.excel:
        metabotk_instance.import_excel(
            file_path=args.excel[0],
            data_sheet=args.excel[1],
        )
    elif args.tables:
        metabotk_instance.import_tables(
            data=args.tables[0],
            sample_metadata=args.tables[1],
            chemical_annotation=args.tables[2],
        )

    ###ANALYZE DATA
    if args.sample_stats:
        sample_stats = metabotk_instance.sample_stats(
            outlier_threshold=args.outlier_threshold,
            exclude_xenobiotics=args.exclude_xenobiotics,
        )
        print(sample_stats.to_csv(sep="\t"))

    if args.metabolite_stats:
        metabolite_stats = metabotk_instance.metabolite_stats(
            outlier_threshold=args.outlier_threshold
        )
        print(metabolite_stats.to_csv(sep="\t"))

    if args.remove_outliers:
        outliers_removed = metabotk_instance.stats.outlier_handler.remove_outliers(
            data_frame=metabotk_instance.data,
            axis=0,
            threshold=args.outlier_threshold,
        )
        print(outliers_removed.to_csv(sep="\t"))

    ### FIT MODELS
    if args.fit_model:
        if len(args.fit_model) == 2:
            residuals = metabotk_instance.models.get_linear_model_residuals(
                formula=args.fit_model[0], models_path=args.fit_model[1]
            )
        else:
            residuals = metabotk_instance.models.get_linear_model_residuals(
                formula=args.fit_model[0], models_path=None
            )
        print(residuals.to_csv(sep="\t"))

    ### PCA

    if args.PCA:
        pca = metabotk_instance.dimensionality_reduction.get_pca(
            n_components=int(args.PCA), get_pca_object=False
        )
        print(pca.to_csv(sep="\t"))

    ### FEATURE SELECTION:
    if args.fs_method:
        if args.fs_method == "boruta":
            ranking = metabotk_instance.feature_selection.boruta(
                y_column=args.y_column,
                threads=args.threads,
                random_state=args.random_state,
                max_depth=args.max_depth,
                class_weight=args.class_weight,
                n_estimators=args.n_estimators,
                alpha=args.alpha,
                max_iterations=args.max_iterations,
                output_dir=args.output_dir,
            )
            print(ranking.to_csv(sep="\t"))

    ###SPLIT DATA
    if args.split_samples:
        if not args.split_samples[0]:
            raise ValueError("No column specified for splitting data")
        else:
            split_samples_dict = metabotk_instance.split_by_sample_column(
                column=args.split_samples[0]
            )
        outdir = args.split_samples[1]
        create_directory(outdir)
        for k, v in split_samples_dict.items():
            k = k.replace(" ", "_").replace("/", "_").replace("-", "_")
            if args.split_samples[2] == "excel":
                v.save_excel(f"{outdir}/{k}.xlsx")
            else:
                v.save_tables(
                    data_path=f"{outdir}/{k}_data.tsv",
                    chemical_annotation_path=f"{outdir}/{k}_metabolites.tsv",
                    sample_metadata_path=f"{outdir}/{k}_samples.tsv",
                )

    if args.split_metabolites:
        if not args.split_metabolites[0]:
            raise ValueError("No column specified for splitting data")
        else:
            split_metabolites_dict = metabotk_instance.split_by_metabolite_column(
                column=args.split_metabolites[0]
            )
        outdir = args.split_metabolites[1]
        create_directory(outdir)
        for k, v in split_metabolites_dict.items():
            k = k.replace(" ", "_").replace("/", "_").replace("-", "_")
            if args.split_metabolites[2] == "excel":
                v.save_excel(f"{outdir}/{k}.xlsx")
            else:
                v.save_tables(
                    data_path=f"{outdir}/{k}_data.tsv",
                    chemical_annotation_path=f"{outdir}/{k}_metabolites.tsv",
                    sample_metadata_path=f"{outdir}/{k}_samples.tsv",
                )
    if args.stratified_kfold:
        metabotk_instance.feature_selection.stratified_kfold(
            n_splits=args.stratified_kfold[0],
            stratification_column=args.stratified_kfold[1],
            output_dir=args.stratified_kfold[2],
        )

    ###SAVE DATA

    if args.output_excel:
        if os.path.exists(args.output_excel[0]):
            metabotk_instance.add_to_excel(
                file_path=args.output_excel[0],
                new_sheet=metabotk_instance.data,
                new_sheet_name=args.output_excel[1],
            )
        else:
            metabotk_instance.save_excel(
                file_path=args.output_excel[0], data_sheet=args.output_excel[1]
            )

    if args.output_tables:
        metabotk_instance.save_tables(
            args.output_tables[0], args.output_tables[1], args.output_tables[2]
        )


if __name__ == "__main__":
    cli()
